#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, time, urlparse, optparse, base64

from ZSI.TC import _get_type_definition, _get_global_element_declaration,\
    b64decode, b64encode, ElementDeclaration
from pyGridWare.utility.scripts.client import GetOptParser, GetPortKWArgs, SetUp
from ConfigParser import ConfigParser

from pyGlobus.security import GSICredentialAttr, GSICredential, \
    STACK_X509_Factory, STACK_X509

from pyGridWare.security.Utility import MessageSigUtility
from pyGridWare.security.gss.GssSecureConversation import GssSecurityContext
from pyGridWare.generated.stubs.ManagedJobFactory_services import *
import pyGridWare.generated.stubs.DelegationFactoryService_services as DELEGATION
import pyGridWare.generated.types.DelegationFactoryService_services_types as DELEGATION_TYPES
import pyGridWare.generated.types.ManagedJobFactory_services_types as MJF_TYPES


#######################################################################
# Couple GED's 
#
#######################################################################

for what in createManagedJobInputMessage.typecode.ofwhat:
    """Runtime patch for Globus.
    """
    if what.pname == 'multiJob':
        what.minOccurs = 0

class CertificateChain_Dec(ElementDeclaration):
    """http://bugzilla.globus.org/bugzilla/show_bug.cgi?id=2496
    Bug in DelegationService ResourceProperty.
    """
    schema = "http://www.globus.org/08/2004/delegationService"
    literal = "CertificateChain"
    def __init__(self, **kw):
        kw["pname"] = ("http://www.globus.org/08/2004/delegationService","CertificateChain")
        kw["aname"] = "_CertificateChain"

        base = _get_type_definition("http://www.globus.org/08/2004/delegationService", "certType")
        if base not in CertificateChain_Dec.__bases__:
            bases = list(CertificateChain_Dec.__bases__)
            bases.insert(0, base)
            CertificateChain_Dec.__bases__ = tuple(bases)
        base.__init__(self, **kw)


class DelegationKey(str, ElementDeclaration):
    '''register delegation key, and specify typecode 
    for serialization/parsing.
    '''
    typecode = ZSI.TC.String(pname=("http://www.globus.org/08/2004/delegationService","DelegationKey"))


#######################################################################
# Functions
#
#######################################################################

def CreateManagedJob(port, EPR, options):
    '''Need to clean up EPR handling, this is mainly an <any> issue,
    need to parse/serialize consistently in/out of <any>

    Essentially copying an EPR obtained from delegation serivce over
    to jobCredentialEndpoint and staggingCredentialEndpoint.

    '''
    key = EPR._ReferenceProperties._any.keys()[0]
    value = EPR._ReferenceProperties._any.values()[0]

    class ResourceKey(str):
        typecode = ZSI.TC.String(pname=key)
    del_key = ResourceKey(value[0])

    request = createManagedJobInputMessage()
    request.__class__.__bases__ = (request.typecode.pyclass,) +  request.__class__.__bases__
    request.typecode.pyclass.__init__(request)

    for what in request.typecode.ofwhat:
        if hasattr(what, 'pyclass') is False: continue
        setattr(request, what.aname, what.pyclass())
        if what.pname == 'job':
            for j in getattr(request, what.aname).typecode.ofwhat:
                if hasattr(j, 'pyclass') is False: continue
                setattr(getattr(request,what.aname), j.aname, j.pyclass())
                 
    request._InitialTerminationTime = time.gmtime(time.time()+20)
    request._JobID = 'uuid:AA108160-47C8-11D9-ABE2-8AF058EA7C2C'
    request._Subscribe = None

    class Holder: pass
    request._job._factoryEndpoint = None
    request._job._jobCredentialEndpoint._Address = EPR._Address
    request._job._jobCredentialEndpoint._ReferenceProperties = Holder()
    request._job._jobCredentialEndpoint._ReferenceProperties._any = del_key

    request._job._stagingCredentialEndpoint._Address = EPR._Address
    request._job._stagingCredentialEndpoint._ReferenceProperties = Holder()
    request._job._stagingCredentialEndpoint._ReferenceProperties._any = del_key
    request._job._executable = options.ex
    request._job._dryRun = False
    request._job._environment = None
    request._job._fileStageIn = None
    request._job._fileStageOut = None
    request._job._fileCleanUp = None

    request._multiJob = None

    return port.createManagedJob(request)


def GetDelegatedCred(options):
    '''swap in path to delegation service, and change to https transport.
    '''
    from copy import deepcopy
    options = deepcopy(options)
    scheme,netloc,path,params,query,fragment = urlparse.urlparse(options.url)
    scheme = 'https'
    path = '/wsrf/services/DelegationFactoryService'
    options.url = urlparse.urlunparse((scheme,netloc,path,params,query,fragment))
    kw = GetPortKWArgs(options)

    loc = DELEGATION.DelegationFactoryServiceLocator()
    port = loc.getDelegationFactoryPortType(**kw)
    req = DELEGATION.GetResourcePropertyRequest(\
        ("http://www.globus.org/08/2004/delegationService","CertificateChain"),
    )
    msg = port.GetResourceProperty(req)

    certChain = msg._any[('http://www.globus.org/08/2004/delegationService', 'CertificateChain')][0]
    stackFact = STACK_X509_Factory()
    bstoken = certChain._any[\
        (u'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd', 
         u'BinarySecurityToken')
    ]
    stack = stackFact.createStackFromEncodedSequence(base64.decodestring(bstoken))
    assert isinstance(stack, STACK_X509), 'stack is wrong type'
    #pubKeyX509 = stack.pop()
    #pubKey = pubKeyX509.get_public_key()
    #verify = pubKey.verify(hash, signature)
    #der = b64encode(stack.get_der_encoding())

    cred = GSICredential()
    cred.read()
    publicKey = cred.get_cert()
    privateKey = cred.get_key()
    credCertChain = cred.get_cert_chain()
    assert isinstance(credCertChain, STACK_X509), 'credCertChain is wrong type'
    credCertChain.insert(publicKey, 0)

    class BST(str):
        typecode = _get_global_element_declaration(\
            "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd", 
            "BinarySecurityToken")(textprotect=False)
        def __new__(cls, value):
            s = str.__new__(cls, value)
            s._attrs = {}
            return s

    der = base64.encodestring(credCertChain.get_der_encoding())
    bst = BST(der)
    rstokMsg = DELEGATION.RequestSecurityTokenMsg()
    rstokMsg._any = [bst,]
    bst._attrs['EncodingType'] =\
        "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
    bst._attrs['ValueType'] =\
         "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-x509-token-profile-1.0#X509PKIPathv1"
    rstokMsgR = port.RequestSecurityToken(rstokMsg)
    return rstokMsgR


def main():
    parser = GetOptParser('ManagedJobFactoryService')
    parser.add_option("-x", "--exec", action="store", dest="ex", help="executable to run.")
    (options, args) = parser.parse_args()
    if not options.ex:
        parser.error("An executable is required")

    SetUp(options)
    kw = GetPortKWArgs(options)

    epr_def_class = _get_type_definition(\
        "http://schemas.xmlsoap.org/ws/2004/03/addressing", \
        "EndpointReferenceType",
    )
    kw['endPointReference'] = epr_def_class(None).pyclass()

    ws_gram = 'http://www.globus.org/namespaces/2004/10/gram/job'
    class Holder: pass
    class ResourceID(str):
        typecode = ZSI.TC.String(pname=(ws_gram,'ResourceID',))

    kw['endPointReference']._ReferenceProperties = Holder()
    kw['endPointReference']._ReferenceProperties._any = {\
        (ResourceID.typecode.nspname,ResourceID.typecode.pname):ResourceID('Fork'),
    }

    loc = ManagedJobFactoryServiceLocator()
    port = loc.getManagedJobFactoryPortType(**kw)

    # GetMultipleResourceProperties
    """
    req = GetMultipleResourcePropertiesRequest()
    req.__class__.__bases__ = req.__class__.__bases__ + (req.typecode.pyclass,)
    req.typecode.pyclass.__init__(req)

    class QName(tuple): pass
    req._ResourceProperty = [\
       QName((ws_gram, 'delegationFactoryEndpoint')),
       QName((ws_gram, 'stagingDelegationFactoryEndpoint')),
    ]
    msg = port.GetMultipleResourceProperties(req)
    """

    rstokMsgR = GetDelegatedCred(options)

    # dEPR -- EPR of delegated credential.
    dEPR = rstokMsgR._any[(u'http://schemas.xmlsoap.org/ws/2004/03/addressing',u'EndpointReference')][0]
    msg = CreateManagedJob(port, dEPR, options)


if __name__ == '__main__':
    main()

