#!/usr/bin/env python
############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, time, urlparse, optparse
from twisted.python import log
from twisted.internet import reactor
from ConfigParser import ConfigParser

import ZSI
import pyGridWare.generated.stubs.counterservice as COUNTER
from pyGridWare.notification.NotificationConsumer import NotificationConsumer
from pyGridWare.utility.scripts.client import GetOptParser, GetPortKWArgs, SetUp


###########################################################################
# Functions
###########################################################################

"""
Counter
    add
    createCounter
"""
def create(port):
    request = COUNTER.CreateCounterRequest()
    msg = port.createCounter(request)
    print "----- create ----- "
    print "\tAddress -- ", msg._EndpointReference._Address
    print "\tRefProp -- ", msg._EndpointReference._ReferenceProperties
    print "\t\tany -- ", msg._EndpointReference._ReferenceProperties._any
    return msg

def add(port, value=999):
    print "----- add ----- "
    request = COUNTER.AddInputMessage(value)
    msg = port.add(request)
    print "RESPONSE: ", msg


"""
ResourceLifetime
    Destroy
    GetCurrentMessage
    SetTerminationTime
"""
def Destroy(port):
    print "----- destroy ----- "
    request = COUNTER.DestroyRequest()
    msg = port.Destroy(request)
    print "RESPONSE: ", msg

def SetTerminationTime(port, interval):
    print "----- SetTerminationTime ----- "
    request = COUNTER.SetTerminationTimeRequest()
    request._RequestedTerminationTime = time.gmtime(time.time()+interval)
    msg = port.SetTerminationTime(request)
    print "TerminationTime: ", msg._NewTerminationTime
    print "CurrrentTime:    ", msg._CurrentTime


"""
ResourceProperties
    GetResourceProperty
"""
def GetResourceProperty(port):
    print "----- GetResourceProperty ----- "
    pyclass = COUNTER_SERVICE.GetResourcePropertyRequest
    request = pyclass(("http://counter.com", "Value"))
    msg = port.GetResourceProperty(request)
    print "RESPONSE: ", msg._any


def Subscribe(port, consumer):
    """
    port -- Subscription Client
    consumer -- Notification Consumer
    Need to specify the prefix-namespace mapping in the
    XML instance to deal with QNames passed via the 
    TopicExpression mixed text content. Do this by using
    the QName TypeCode.
    """
    print "----- Subscribe ----- "

    class TopicExpression:
        class QName(tuple):
            typecode = ZSI.TC.QName()

        def __init__(self, namespaceURI, name):
            self._attrs = {}
            self._attrs['Dialect'] = \
                "http://docs.oasis-open.org/wsn/2004/06/TopicExpression/Simple"
            self._text = TopicExpression.QName((namespaceURI,name))

    
    request = COUNTER.SubscribeRequest()
    epr = consumer.context.getConsumerReference()
    request._ConsumerReference = epr
    request._TopicExpression = TopicExpression("http://counter.com", "Value")
    request._UseNotify = True
    msg = port.Subscribe(request)
    print "RESPONSE: ", msg


def main():
    op = GetOptParser('CounterService')
    (options, args) = op.parse_args()
    SetUp(options)
    kw = GetPortKWArgs(options)

    locator = COUNTER.CounterServiceLocator()
    port = locator.getCounterPortType(**kw)

    response = create(port)

    # Get New port w/ EPR
    kw['endPointReference'] = response._EndpointReference
    port = locator.getCounterPortType(**kw)

    consumer = NotificationConsumer()
    consumer.start()

    print "=======> Queue Events.."
    reactor.callWhenRunning(Subscribe, port, consumer)
    reactor.callWhenRunning(add, port, 13)
    reactor.callLater(4, Destroy, port)
    reactor.callLater(5, reactor.stop)

 
if __name__ == '__main__':
    reactor.callWhenRunning(main)
    reactor.run()



