############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import sys, os, time
import ZSI.TC 
from ZSI.TC import _get_type_definition as GTD
from ZSI.wstools.Namespaces import OASIS
from twisted.internet import reactor
from twisted.python import log
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.properties.ResourceProperties import NotificationProducerRP
from pyGridWare.generated.stubs.basenotification import \
        Notify, NotificationConsumerSOAPBindingSOAP, WS_BaseNotificationServiceLocator
import pyGridWare.generated.types.basenotification
import pyGridWare.generated.types.notificationconsumerservice

'''
'''

MESSAGE = GTD("http://wsrf.globus.org/core/notification",
    "ResourcePropertyValueChangeNotificationElementType")(pname='Message')
RPV = filter(\
    lambda w: (w.nspname,w.pname) == (OASIS.PROPERTIES,'ResourcePropertyValueChangeNotification'),
    MESSAGE.ofwhat,
    )[0]
NEWVALUE = filter(\
    lambda w: (w.nspname,w.pname) == (OASIS.PROPERTIES,'NewValue'),
    RPV.ofwhat, 
    )[0]

def Message(property):
    pyobj = MESSAGE.pyclass()
    rpv = pyobj._ResourcePropertyValueChangeNotification = RPV.pyclass()
    rpv._NewValue = NEWVALUE.pyclass()
    rpv._NewValue._any = property
    return pyobj


class NotificationProducer:
    '''The NotificationProducer can act as a SubscriptionManager, 
    request -- subscription request, containing ConsumerReference and TopicExpression.
    ctx -- resource context subscription applies to.
    '''
    subscriptions = []

    def __init__(self):
        '''
        stub -- NotificationConsumer RPC instance
        subscription -- Subscription instance
        '''
        self.stub = None
        self.subscription = None

    def setUp(self, subscription, **kw):
        '''
        request -- ConsumerReference element declaration, contents
            ConsumerReference, TopicExpression, UseNotify
        '''
        producer = subscription.producer
        properties = producer.properties 
        request = subscription.request

        if isinstance(properties, NotificationProducerRP) is False:
            raise SubscribeCreationFailedFault(timestamp=time.gmtime(),
                faultcause=('Invalid Notification Producer Properties.'),
            )

        url = subscription.reference._Address
        locator = WS_BaseNotificationServiceLocator()

        self.stub = locator.getNotificationConsumer(url=url, \
            endPointReference=request._ConsumerReference, tracefile=sys.stdout, **kw)

        self.subscription = subscription
        properties.addTopicListener(self.Notify, subscription.topic)


    class Holder: 
        pass

    class QName(tuple):
        typecode = ZSI.TC.QName()

    def Notify(self, property, topicExpression):
        '''Send Notify message.
        '''
        log.debug("Send Notification -->")
        request = Notify()

        NotificationMessage = NotificationProducer.Holder()
        request._NotificationMessage = (NotificationMessage,)

        NotificationMessage._Topic = NotificationProducer.Holder()
        NotificationMessage._Topic._text = \
            NotificationProducer.QName((property.typecode.nspname,property.typecode.pname))

        NotificationMessage._ProducerReference = \
            AddressingUtils.createEndpointReference(self.subscription.producer)
        NotificationMessage._Message = Message(property)
        #NotificationMessage._Message._ResourcePropertyValueChangeNotification._NewValue._any = \
        #    property

        self.stub.Notify(request)

