############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
from ZSI.TC import _get_global_element_declaration as GED
from twisted.python import log
from pyGridWare.patterns import Singleton
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.notification.NotificationProducer import NotificationProducer
from pyGridWare.resource.ResourceContext import ResourceContext
from pyGridWare.lifetime.ResourceLifetime import ImmediateResourceTermination,\
    ScheduledResourceTermination

import pyGridWare.generated.types.basenotification
from pyGridWare.generated.stubs.basenotification import \
    WS_BaseNotificationServiceLocator, NotificationConsumerSOAPBindingSOAP

from pyGridWare.generated.resource.submanagerservice.SubscriptionManagerService import \
    SubscriptionManagerContext


CONSUMER_REF = GED(\
    "http://docs.oasis-open.org/wsn/2004/06/wsn-WS-BaseNotification-1.2-draft-01.xsd",
    "ConsumerReference",
)


class Subscription(object):
    '''Represents a subscription, registered to NotificationProducer
    '''
    def __init__(self, context, producer):
        '''
        context -- SubscriptionManagerContext instance.
        producer -- Monitoring Resource Context 
        '''
        self.context = context
        self.producer = producer
        self.request = None
        self.topic = None

    def getContext(self):
        return self.__sub_context
    def setContext(self, context):
        assert isinstance(context, SubscriptionManagerContext), \
            'resource must be a SubscriptionManagerContext instance'
        self.__sub_context = context
    context = property(getContext, setContext, None,
        'Subscription Context')
 
    def getProducerContext(self):
        return self.__p_context
    def setProducerContext(self, context):
        assert isinstance(context, ResourceContext), \
            'resource must be a ResourceContext instance'
        self.__p_context = context
    producer = property(getProducerContext, setProducerContext, None,
        'Subscribed to Resource Context')

    def getSubscriptionRequest(self):
        return self.__request
    def setSubscriptionRequest(self, request):
        self.__request = request
    request = property(getSubscriptionRequest, setSubscriptionRequest, None,
        'Subscription Request')

    def getTopicExpression(self):
        return self.__topic
    def setTopicExpression(self, topic):
        assert topic is None or (hasattr(topic, '_attrs') and topic._attrs.has_key('Dialect')), \
            'topic must have Dialect attribute'
        self.__topic = topic
    topic = property(getTopicExpression, setTopicExpression, None,
        'TopicExpression')

    def getConsumerReference(self):
        return self.__reference
    def setConsumerReference(self, reference):
        self.__reference = reference
    reference = property(getConsumerReference, setConsumerReference, None,
            'Consumer EPR (%s,%s)' %(CONSUMER_REF.schema, CONSUMER_REF.literal))

    def setUp(self, request, **kw):
        '''
        request -- ConsumerReference element declaration, contents
            ConsumerReference, TopicExpression, UseNotify
        '''
        self.request = request
        if self.request._UseNotify is True:
            self.reference = self.request._ConsumerReference
            self.topic = self.request._TopicExpression
            self.notificationProducer = NotificationProducer()
            self.notificationProducer.setUp(self, **kw)



