############################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import ZSI.TC 
from ZSI.TC import _get_global_element_declaration as GED
from ZSI.TC import ElementDeclaration
from pyGridWare.generated.services.secureconversation.SecureConversationService import \
    SecureConversationServiceWSRF as WSRFService
from pyGridWare.security.gss.GssSecurityContext import GssSecurityContext
from ZSI.wstools.Namespaces import WSSE, BEA
import pyGridWare.generated.types.secureconversation


#Move this somewhere else, the AuthenticationHandler and the SecureConversation
#need these classes
class Holder:
    pass

class RequestType(str):
    typecode = GED("http://schemas.xmlsoap.org/ws/2004/04/trust","RequestType")()

class Identifier:
    typecode =  GED("http://schemas.xmlsoap.org/ws/2002/07/utility","Identifier")()
    
class ContextAddress:
    typecode = GED("http://wsrf.globus.org/core/2004/07/security/secconv","ContextAddress")()

class SecurityContextToken:
    typecode = GED("http://schemas.xmlsoap.org/ws/2004/04/sc","SecurityContextToken")()

class SecureKey(str):
    typecode = ZSI.TC.String(pname=("http://www.globus.org/namespaces/2004/06/core","SecureContextKey "))

class Token(str):
    typecode = GED("http://schemas.xmlsoap.org/ws/2004/04/trust","TokenType")()

class BinaryExchange(str):
    typecode = GED("http://schemas.xmlsoap.org/ws/2004/04/trust","BinaryExchange")()
    def __new__(cls, value):
        s = str.__new__(cls, value)
        s._attrs = {}
        return s

class SecureConversationService(WSRFService):
    
    def __init__(self, post='/wsrf/services/', **kw):
        WSRFService.__init__(self, post)

    def wsa_RequestSecurityToken(self, ps, address):
        response = WSRFService.wsa_RequestSecurityToken(self, ps, address)
        context = GssSecurityContext()

        for any in self.request._any:
            if hasattr(any, 'typecode') and \
                (any.typecode.nspname,any.typecode.pname) == (WSSE.TRUST,"BinaryExchange"):
                break
        else:
            raise RuntimeError, \
                '%s RequestSecurityToken did not receive a BinaryExchange token' %self.__class__

        tokens = any
        outTokens = context.acceptSecurityContext(tokens)
        #contextID = context.hashCode()
        home = GssSecurityContext.getResourceHome()
        contextID = home.create()
        #home.create(context, contextID)
        home.add(context, contextID)

        response._any = []
        response._attrs = {}
        #response._attrs['Context'] = 'uuid:' + contextID
        response._attrs['Context'] = 'uuid:%d' %contextID
        response._any.append(RequestType('http://schemas.xmlsoap.org/ws/2004/04/security/trust/Issue'))
        response._any.append(Token('http://schemas.xmlsoap.org/ws/2004/04/security/sc/sct'))
        
        binEx = BinaryExchange(outTokens)
        response._any.append(binEx)
        response._any[-1]._attrs['EncodingType'] = \
                                                "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
        response._any[-1]._attrs['ValueType'] = \
                                             "http://www.globus.org/ws/2004/09/security/sc#GSSAPI_GSI_TOKEN"                     
        return response

    def wsa_RequestSecurityTokenResponse(self, ps, address):
        response = WSRFService.wsa_RequestSecurityTokenResponse(self, ps, address)

        for any in self.request._any:
            if hasattr(any, 'typecode') and \
                (any.typecode.nspname,any.typecode.pname) == (WSSE.TRUST,"BinaryExchange"):
                break
        else:
            raise RuntimeError, \
                '%s RequestSecurityTokenResponse did not receive a BinaryExchange token' %self.__class__

        tokens = any
        contextID = self.request._attrs['Context']
        home = GssSecurityContext.getResourceHome()
        context_id = contextID.split(':')[1]
        context = home.get(context_id)
        outTokens = context.acceptSecurityContext(tokens)
        response._attrs = {}
        response._attrs['Context'] =  contextID
        response._any = []
        if context.contextCreated():
            response._any.append(SecurityContextToken())
            response._any[0]._Identifier = contextID
            response._any[0]._any = [ContextAddress()]
            response._any[0]._any[0]._Address = address._addressTo
            response._any[0]._any[0]._ReferenceProperties = Holder()
            response._any[0]._any[0]._ReferenceProperties._any = [SecureKey(context_id)]
        else:
            binEx = BinaryExchange(outTokens)
            response._any.append(binEx)
            response._any[-1]._attrs['EncodingType'] = \
                "http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-soap-message-security-1.0#Base64Binary"
            response._any[-1]._attrs['ValueType'] = \
                                          "http://www.globus.org/ws/2004/09/security/sc#GSSAPI_GSI_TOKEN"
        return response
