############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
import time
from pyGridWare.addressing.AddressingUtils import \
    AddressingUtils
from pyGridWare.lifetime.ResourceLifetime import \
    ImmediateResourceTermination, ScheduledResourceTermination
from pyGridWare.services.SubscriptionManager import \
    SubscriptionManager
from pyGridWare.security.gss.GssService import GssSecureResourceMixIn
from pyGridWare.generated.services.performancecounter.CounterService import CounterServiceWSRF as WSRFService



class Counter(WSRFService, GssSecureResourceMixIn):

    def __init__(self, post='/wsrf/services/CounterService', **kw):
        WSRFService.__init__(self, post, **kw)
        GssSecureResourceMixIn.__init__(self, post)

    def wsa_GetCurrentMessage(self, ps, address):
        # Get request and response
        response = WSRFService.wsa_GetCurrentMessage(self, ps, address)
        self.request

        # Get Resource Context
        ctx = WSRFService.GetResourceContext(ps, address)

        return response

    def wsa_Subscribe(self, ps, address):
        # Get request and response
        response = WSRFService.wsa_Subscribe(self, ps, address)
        self.request

        # Get Resource Context
        ctx = WSRFService.GetResourceContext(ps, address)

        # Pass SubscriptionRequest off
        assert SubscriptionManager._isInstantiated(), 'SubscriptionManager is not instantiated'
        manager = SubscriptionManager.getInstance()
        response._SubscriptionReference = manager.subscribe(self.request, ctx)
        return response

    def wsa_add(self, ps, address):
        # Get request and response
        response = WSRFService.wsa_add(self, ps, address)
        self.request

        # Get Resource Context
        ctx = WSRFService.GetResourceContext(ps, address)

        # Access resource properties, and evaluate new value.
        counter = ctx.properties
        result = counter.Value
        result += self.request
        counter.Value = result

        # response is immutable instance, create new one with result.
        response = response.__class__(result)
        return response

    def wsa_createCounter(self, ps, address):
        # Get request and response
        response = WSRFService.wsa_createCounter(self, ps, address)
        self.request

        # Create Resource Context
        ctx = self.NewResourceContextCounterPortTypePort(ps, address)

        # Initialize Value to 0
        counter = ctx.properties
        counter.Value = 0

        # Create EndPointReference and set in response
        epr = AddressingUtils.createEndpointReference(ctx)
        response._EndpointReference = epr
        return response


