############################################################################
# Joshua R. Boverhof, LBNL
# See LBNLCopyright for copyright notice!
###########################################################################
from twisted.python import log
from pyGridWare.patterns import Singleton
from pyGridWare.addressing.AddressingUtils import AddressingUtils
from pyGridWare.resource.ResourceContext import ResourceContext
from pyGridWare.lifetime.ResourceLifetime import ImmediateResourceTermination,\
    ScheduledResourceTermination
from pyGridWare.notification.SubscriptionManager import Subscription

import pyGridWare.generated.types.basenotification
from pyGridWare.generated.resource.submanagerservice.SubscriptionManagerService import \
     SubscriptionManagerContext
from pyGridWare.generated.services.stubs.submanagerservice import \
    SubscriptionManagerService
from ZSI.wstools.Namespaces import WSA200403


class SubscriptionManager(SubscriptionManagerService, Singleton):
    '''
    service -- SubscriptionManagerService.
    subscriptions -- list of Subscriptions.
    '''
    service = None
    subscriptions = []
    kw = {\
        'wsAddressURI':WSA200403.ADDRESS,
    }
    def __init__(self, post='/wsrf/services/SubscriptionManagerService', **kw):
        SubscriptionManagerService.__init__(self, post, **kw)

    def NewSubscription(cls, request, ctx, subCtx):
        '''Create a new subscription
        request -- subscription request, containing ConsumerReference and TopicExpression.
        ctx -- resource context subscription applies to.
        subCtx -- subscription resource context.
        '''
        sub = Subscription(subCtx, ctx)
        sub.setUp(request, **SubscriptionManager.kw)
        cls.subscriptions.append(sub)
    NewSubscription = classmethod(NewSubscription)
 
    def Destroy(ps, address):
        home = SubscriptionManagerContext.getResourceHome()
        ctx = home.getResourceContext(ps, address)
        ImmediateResourceTermination.Destroy(ctx)
    Destroy = staticmethod(Destroy)

    def subscribe(self, request, ctx):
        '''Returns an EPR representing a SubscriptionReference.
        request -- subscription request, containing ConsumerReference and TopicExpression.
        ctx -- resource context subscription applies to.
        '''
        assert isinstance(ctx, ResourceContext), \
            'resource must be a ResourceContext instance'

        # Subscription Context
        sub_ctx = SubscriptionManagerContext()
        sub_ctx.setAddress(self.getServiceURL())
        sub_home = sub_ctx.getResourceHome()
        sub_key = sub_home.create()
        sub_home.add(sub_ctx, sub_key)
        sub_ctx.setResourceID(sub_key)

        self.NewSubscription(request, ctx, sub_ctx)
        epr = AddressingUtils.createEndpointReference(sub_ctx)
        return epr

    # Callback methods
    def wsa_Destroy(self, ps, address):
        response = SubscriptionManagerService.wsa_Destroy(self, ps, address)
        self.Destroy(ps, address)
        return response


