###########################################################################
# Monte M. Goode, LBNL
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################

# $Id: resource.py,v 1.20 2005/04/28 05:47:56 boverhof Exp $

import os,sys,warnings
from cStringIO import StringIO
from twisted import copyright
TWISTED_VERSION = (2,0,0)
try:
    if tuple(map(lambda i:int(i), copyright.version.split('.'))) < TWISTED_VERSION:
        warnings.warn(\
            'Twisted-%s or later is required, found: %s' %(TWISTED_VERSION, copyright.version),
            RuntimeWarning,
        )
except:
    warnings.warn(\
        'Twisted-%s or later is required, found: %s' %(TWISTED_VERSION, copyright.version),
        RuntimeWarning,
    )

from twisted.web.error import NoResource
from twisted.web.server import NOT_DONE_YET
from twisted.web.resource import Resource
from twisted.python import log, failure

from ZSI import EvaluateException, ParseException
from ZSI.parse import ParsedSoap
from ZSI.writer import SoapWriter
from ZSI import fault 
from ZSI.address import Address
from ZSI.ServiceContainer import WSActionException, ServiceInterface,\
    PostNotSpecified, SOAPActionNotSpecified, WSActionNotSpecified

try:
    from pyGridWare.utility.dom import DomletteReader, DomletteElementProxy
except ImportError, ex: 
    DomletteReader = DomletteElementProxy = None
    
from pyGridWare.faults import BaseFault

class FileServerResource(Resource, object):
    '''Represents a root directory to serve files out of.
    '''
    def __init__(self, root='wsdl'):
        '''root -- root directory path.
        '''
        self.root = root
        Resource.__init__(self)

    def render_GET(self, request):
        print "FIND %s " %request.path
        s = StringIO()
        if request.path.startswith('/%s' %self.root) is False:
            raise NoResource, request.path

        rpath = request.path.split(self.root, 1)[1].lstrip('/')
        fpath = os.path.join(self.root, rpath)
        if os.path.isfile(fpath) is True:
            for l in open(fpath,'r'): s.write(l)
            return s.getvalue()

        raise NoResource, 'request(%s), no file(%s)' %(request.path,fpath)

    def getChild(self, path, request):
        '''so can serve GET requests and act as directory.
        '''
        return self


class WSResource(Resource, ServiceInterface, object):
    '''
    class variables:
        encoding  
        soapAction -- Must override to set input Action values.
           Dictionary of soapAction keys, and operation name values.
           These are specified in the WSDL soap bindings. There must be a 
           class method matching the operation name value.  If WS-Action is
           used the keys are WS-Action request values, according to the spec
           if soapAction and WS-Action is specified they must be equal.
           
        wsAction -- Must override to set output Action values.
           Dictionary of operation name keys and WS-Action 
           response values.  These values are specified by the portType.

        rootDict -- dictionary of root element keys, and operation name values.
    '''
    encoding = "UTF-8"

    def __init__(self, post=None):
        '''
        post -- POST value, also have access via Resource.
        '''
        assert isinstance(self.soapAction, dict), \
            "soapAction must be a dict not: %s" %type(self.soapAction)
        assert isinstance(self.wsAction, dict), \
            "wsAction must be a dict not: %s" %type(self.wsAction)

        self.url = None
        ServiceInterface.__init__(self, post)
        Resource.__init__(self)

    def getServiceURL(self):
        return self._url
    def setServiceURL(self, url):
        self._url = url
    url = property(getServiceURL, setServiceURL, None, 'URL of service')

    def getOperation(self, ps, address):
        '''Returns a method of class.
        address -- ws-address 
        '''
        action = address.getAction()
        opName = self.getOperationName(ps, action)
        return getattr(self, opName)

    def getResponseAction(self, ps, action):
        '''Returns response WS-Action if available
           action -- request WS-Action value.
        '''
        opName = self.getOperationName(ps, action)
        if self.wsAction.has_key(opName) is False:
            raise WSActionNotSpecified, 'wsAction dictionary missing key(%s)' %opName
        return self.wsAction[opName]

    def render_POST(self, request):
        """ Method called by twisted, not by HTTPServer.
        request -- twisted.web.server.Request
        """
        if request.path != self.post:
            raise Warning, \
                'The request path(%s) doesn\'t match the service post(%s)' \
                %(request.path,self.post)

        http_headers = request.getAllHeaders()
        soap_action = http_headers.get('soapaction')
        log.debug('Request soapAction -- %s' %soap_action)
        data = request.content.read()
        log.debug(data)
        try:
            ps = ParsedSoap(data, readerclass=DomletteReader)
        except ParseException, e:
            return self._soapFault(request, e)

        # Verify if Signed
        self.verify(ps)

        # WS-Address
        address = Address()
        netloc = '%s:%d' %(request.host.host, request.host.port)
        if request.transport.TLS == 0:
            localURL = 'http://%s%s' %(netloc, request.path)
        else:
            localURL = 'https://%s%s' %(netloc, request.path)

        log.debug('Service local URL: %s' %localURL)
        self.setServiceURL(localURL)
        try:
            address.parse(ps)
        except ParseException, e:
            return self._soapFault(request, e)

        action = address.getAction()
        if not action:
            log.err('No WS-Action specified in Request')
            e = WSActionException('No WS-Action specified in Request')
            return self._soapFault(request, e)

        if soap_action and soap_action.strip('\'"') != action:
            log.err('SOAP Action("%s") must match WS-Action("%s") if specified.' %(soap_action,action))
            e = WSActionException(\
                'SOAP Action("%s") must match WS-Action("%s") if specified.' %(soap_action,action)
            )
            return self._soapFault(request, e)

        try:
            bound_method = self.getOperation(ps, address)
            result = bound_method(ps, address)
        except Exception, ex:
            log.err('Failed to locate service instance/bound method from WS-Address')
            return self._soapFault(request, ex)

        if result is None:
            log.debug('No Response specified')
            return NOT_DONE_YET

        sw = SoapWriter(outputclass=DomletteElementProxy)
        try:
            sw.serialize(result)
        except Exception, ex:
            return self._soapFault(request, ex)

        rsp_action = self.getResponseAction(ps, action)
        addressRsp = Address(action=rsp_action)
        addressRsp.setResponseFromWSAddress(address, localURL)
        try:
            addressRsp.serialize(sw)
        except Exception, ex:
            log.err('Problem serializing WS-Address for Response')
            return self._soapFault(request, e)

        self.sign(sw)
        soapdata = str(sw)
        self._sendResponse(request, soapdata)
        return NOT_DONE_YET

    def _sendResponse(self, request, response, status=200):
        request.setResponseCode(status)

        if self.encoding is not None:
            mimeType = 'text/xml; charset="%s"' % self.encoding
        else:
            mimeType = "text/xml"

        request.setHeader("Content-type", mimeType)
        request.setHeader("Content-length", str(len(response)))
        request.write(response)
        request.finish()

    def _soapFault(self, request, e):
        response = None
        if isinstance(e, BaseFault):
            response = fault.Fault(fault.Fault.Server, 'userException', None, [e.fault], None).AsSoap()
        else:
            response = fault.FaultFromZSIException(e).AsSOAP()

        log.err('SOAP FAULT: %s' % response)
        self._sendResponse(request, response, status=500)
        return NOT_DONE_YET

