##########################################################################
# Joshua R. Boverhof, LBNL
# See Copyright for copyright notice!
###########################################################################
import os, sys
from distutils.core import setup
from distutils.dist import Distribution
from os.path import isfile
from ConfigParser import ConfigParser
from ZSI.wstools.WSDLTools import WSDLReader

sys.path.append('%s/pyGridWare/utility/generate' %os.getcwd())
from Modules import WriteServiceModules, WriteClientStubModule,\
    WriteTypesModule

CONFIG_FILE = 'config.txt'
CACHE_FILE = '.pyGridWare.cache'


def RegenerateModules(items, services=True, stubs=True):
    '''Regenerate generated modules with new genrator
    '''
    if services is stubs is False:
        return

    add_packages = []
    for sec,url in items:
        print "Generate: ",sec,url
        wsdl = None
        reader = WSDLReader()
        if isfile(url) is False:
            wsdl = reader.loadFromURL(url)
        else:
            wsdl = reader.loadFromFile(url)

        if services:
            WriteServiceModules(sec, wsdl, add_packages)
        if stubs:
            WriteClientStubModule(sec, wsdl)
        else:
            WriteTypesModule(sec, wsdl)

    return add_packages


class PyGridWareDist(Distribution): 
    global_options = Distribution.global_options[:]
    global_options.append(('regenerate', None, "regenerate modules in pyGridWare.generated"))

    def parse_command_line(self):
        ok = Distribution.parse_command_line(self)
        if self.verbose == 1:
            from ZSI.wstools.logging import setBasicLoggerDEBUG
            setBasicLoggerDEBUG()

        cp = ConfigParser()
        if '--regenerate' in self.script_args:
            cp.read(CONFIG_FILE)
            packages = RegenerateModules(cp.items('WSDL'), services=True, stubs=True)
            self.packages += packages
            fd = open(CACHE_FILE, 'w+')
            fd.write('[distribution]\npackages = %s' %packages)
        else:
            cp.read(CACHE_FILE)
            self.packages += eval(cp.get('distribution', 'packages'))

        del cp
        return ok 


def main():
    classifiers = """\
        Development Status :: 5 - Production/Stable
        Intended Audience :: Developers
        Intended Audience :: Science/Research
        License :: OSI Approved :: BSD License
        Programming Language :: Python
        Natural Language :: English
        Operating System :: Microsoft :: Windows
        Operating System :: POSIX
        """

    longDesc = """
        Python implementation of the WS-Resource Framework (WSRF).  
        pyGridWare has facilities for secure messaging including security 
        context negotiation, delegation, and GSS XML Signatures.
        """

    setup(name="pyGridWare", 
          version="1.0.2beta", 
          packages=["pyGridWare", 
                    "pyGridWare.addressing", 
                    "pyGridWare.generated", 
                    "pyGridWare.generated.properties", 
                    "pyGridWare.generated.resource", 
                    "pyGridWare.generated.services", 
                    "pyGridWare.generated.services.stubs", 
                    "pyGridWare.generated.stubs", 
                    "pyGridWare.generated.types", 
                    "pyGridWare.lifetime", 
                    "pyGridWare.notification", 
                    "pyGridWare.properties",
                    "pyGridWare.faults",
                    "pyGridWare.resource", 
                    "pyGridWare.security", 
                    "pyGridWare.security.gss", 
                    "pyGridWare.security.base",
                    "pyGridWare.services", 
                    "pyGridWare.utility", 
                    "pyGridWare.utility.generate", 
                    "pyGridWare.utility.scripts",
                    "pyGridWare.wsrf", 
          ],
          scripts=["tools/WSDLBuilder"],
          description="Python WS-Resource Framework and GSS Security",
	  author="Joshua Boverhof", 
          author_email="JRBoverhof@lbl.gov",
          maintainer="Joshua Boverhof", 
          maintainer_email="JRBoverhof@lbl.gov",
	  url="http://www-itg.lbl.gov/gtg/pyGridWare/",
          classifiers = filter(None, classifiers.split("\n"))[:-1],
          long_description=longDesc,
          distclass=PyGridWareDist,
    )


if __name__ == '__main__': 
    main()
