#! /usr/bin/env python
# $Header: /cvsroot/pywebsvcs/zsi/ZSI/fault.py,v 1.8.2.11 2005/04/27 21:24:21 boverhof Exp $
'''Faults.
'''

from ZSI import _copyright, _children, _child_elements, \
        _textprotect, _stringtypes, _seqtypes, _Node, SoapWriter

from ZSI.TCcompound import Struct
from ZSI.TC import QName, URI, String, XMLString, AnyElement, UNBOUNDED

from ZSI.wstools.Namespaces import SOAP, ZSI_SCHEMA_URI
from ZSI.wstools.c14n import Canonicalize
from ZSI.TC import ElementDeclaration

import traceback, cStringIO as StringIO


class Detail:
    def __init__(self, any=None):
        self.any = any

Detail.typecode = Struct(Detail, [AnyElement(aname='any',minOccurs=0, maxOccurs="unbounded")], pname='detail', minOccurs=0)

class FaultType:
    def __init__(self, faultcode=None, faultstring=None, faultactor=None, detail=None):
        self.faultcode = faultcode
        self.faultstring= faultstring
        self.faultactor = faultactor
        self.detail = detail
        
FaultType.typecode = \
    Struct(FaultType,
        [QName(pname='faultcode'), 
         String(pname='faultstring'),
         URI(pname=(SOAP.ENV,'faultactor'), minOccurs=0),
         Detail.typecode,
         AnyElement(aname='any',minOccurs=0, maxOccurs=UNBOUNDED),
        ], 
        pname=(SOAP.ENV,'Fault'), 
        inline=True,
        hasextras=0, 
    )

class ZSIHeaderDetail(ElementDeclaration):
    schema = ZSI_SCHEMA_URI
    literal = (ZSI_SCHEMA_URI, 'detail')
    def __init__(self, detail):
        self.any = detail

ZSIHeaderDetail.typecode =\
    Struct(ZSIHeaderDetail, [AnyElement(aname='any',)], pname=(ZSI_SCHEMA_URI, 'detail'))


class ZSIFaultDetail(ElementDeclaration):
    '''<ZSI:FaultDetail>
           <ZSI:string>%s</ZSI:string>
           <ZSI:trace>%s</ZSI:trace>
       </ZSI:FaultDetail>
'''
    schema = ZSI_SCHEMA_URI
    literal = (ZSI_SCHEMA_URI, 'FaultDetail')
    def __init__(self, string=None, trace=None):
        self.string = string
        self.trace = trace

ZSIFaultDetail.typecode =\
    Struct(ZSIFaultDetail, [String(pname=(ZSI_SCHEMA_URI, 'string')),
         String(pname=(ZSI_SCHEMA_URI, 'trace'),minOccurs=0),], 
         pname=(ZSI_SCHEMA_URI, 'FaultDetail'),
    )


class URIFaultDetail(ElementDeclaration):
    '''
    <ZSI:URIFaultDetail>
        <ZSI:URI>uri</ZSI:URI>
        <ZSI:localname>localname</ZSI:localname>
    </ZSI:URIFaultDetail>
    '''
    schema = ZSI_SCHEMA_URI
    literal = (ZSI_SCHEMA_URI, 'URIFaultDetail')
    def __init__(self, uri=None, localname=None):
        self.URI = uri
        self.localname = localname

URIFaultDetail.typecode =\
    Struct(URIFaultDetail, [String(pname=(ZSI_SCHEMA_URI, 'URI')),
        String(pname=(ZSI_SCHEMA_URI, 'localname')),], pname=(ZSI_SCHEMA_URI, 'URIFaultDetail'),
    )

class ActorFaultDetail:
    '''
    <ZSI:ActorFaultDetail>
        <ZSI:URI>%s</ZSI:URI>
    </ZSI:ActorFaultDetail>
    '''
    def __init__(self, uri=None):
        self.URI = uri

ActorFaultDetail.typecode =\
    Struct(ActorFaultDetail, [String(pname=(ZSI_SCHEMA_URI, 'URI')),],
        pname=(ZSI_SCHEMA_URI, 'ActorFaultDetail'),
    )

class Fault(Exception):
    '''SOAP Faults.
    '''

    Client = "SOAP-ENV:Client"
    Server = "SOAP-ENV:Server"
    MU     = "SOAP-ENV:MustUnderstand"

    def __init__(self, code, string,
                actor=None, detail=None, headerdetail=None):
        self.code, self.string, self.actor, self.detail, self.headerdetail = \
                code, string, actor, detail, headerdetail
        Exception.__init__(self, code, string, actor, detail, headerdetail)

    def DataForSOAPHeader(self):
        if not self.headerdetail: return None
        # SOAP spec doesn't say how to encode header fault data.
        return ZSIHeaderDetail(self.headerdetail)

    def serialize(self, sw):
        '''Serialize the object.'''
        detail = None
        if self.detail is not None: 
            detail = Detail()
            detail.any = self.detail

        pyobj = FaultType(self.code, self.string, self.actor, detail)
        sw.serialize(pyobj, typed=False)

    def AsSOAP(self, output=None, **kw):
        if output is not None: raise RuntimeError, "NOT SUPPORTED.."

        header = self.DataForSOAPHeader() 
        sw = SoapWriter(**kw)
        self.serialize(sw)
        if header is not None:
            sw.serialize_header(header, header.typecode, typed=False)
        return str(sw)

    AsSoap = AsSOAP


def FaultFromNotUnderstood(uri, localname, actor=None):
    detail, headerdetail = None, URIFaultDetail(uri, localname)
    return Fault(Fault.MU, 'SOAP mustUnderstand not understood',
                actor, detail, headerdetail)


def FaultFromActor(uri, actor=None):
    detail, headerdetail = None, ActorFaultDetail(uri)
    return Fault(Fault.Client, 'Cannot process specified actor',
                actor, detail, headerdetail)


def FaultFromZSIException(ex, actor=None):
    '''Return a Fault object created from a ZSI exception object.
    '''
    mystr = getattr(ex, 'str', None) or str(ex)
    mytrace = getattr(ex, 'trace', '')
    elt = '''<ZSI:ParseFaultDetail>
<ZSI:string>%s</ZSI:string>
<ZSI:trace>%s</ZSI:trace>
</ZSI:ParseFaultDetail>
''' % (_textprotect(mystr), _textprotect(mytrace))
    if getattr(ex, 'inheader', 0):
        detail, headerdetail = None, elt
    else:
        detail, headerdetail = elt, None
    return Fault(Fault.Client, 'Unparseable message',
                actor, detail, headerdetail)

def FaultFromException(ex, inheader, tb=None, actor=None):
    '''Return a Fault object created from a Python exception.
    '''
    tracetext = None
    if tb:
        try:
            lines = '\n'.join(['%s:%d:%s' % (name, line, func)
                        for name, line, func, text in traceback.extract_tb(tb)])
        except:
            pass
        else:
            tracetext = lines
  
    elt = ZSIFaultDetail(string=str(ex), trace=tracetext)
    if inheader:
        detail, headerdetail = None, (elt,)
    else:
        detail, headerdetail = (elt,), None
    return Fault(Fault.Server, 'Processing Failure',
                actor, detail, headerdetail)

def FaultFromFaultMessage(ps):
    '''Parse the message as a fault.
    '''
    pyobj = ps.Parse(FaultType.typecode)
    return Fault(pyobj.faultcode, pyobj.faultstring,
                pyobj.faultactor, pyobj.detail.any)


if __name__ == '__main__': print _copyright
